@@ -1,30 +1,30 @@
 #!/usr/bin/env python
 
-import os
-import sys
-import re
-
-import subprocess
-import tempfile
-import shutil
-import storage
-import gzip
-from io import BytesIO
-import rpmfile
-import hashlib
-import json
-import itertools
-
 import datetime
+import gzip
+import hashlib
+import os
+import re
+import shutil
+import subprocess
+import sys
+import tempfile
 import time
-
+from io import BytesIO
 from xml.sax.saxutils import escape
 
+import rpmfile
+import storage
+
 try:
     import xml.etree.cElementTree as ET
 except ImportError:
     import xml.etree.ElementTree as ET
 
+# Changelog limit is used to get only last CHANGELOG_LIMIT changelog lines.
+# Usually it takes last 10.
+CHANGELOG_LIMIT = 10
+
 
 def gzip_bytes(data):
     out = BytesIO()
@@ -58,6 +58,18 @@ def bytes_checksum(data, checksum_type):
 
 
 def gpg_sign_string(data, keyname=None, inline=False):
+    """Signing data according to the specified options.
+
+    Keyword arguments:
+    data - data for sign (Unicode string).
+    keyname - name of the gpg key that will be used to sign the
+              data (string, default: None).
+    inline - option specifies whether to use a cleartext
+             signature (bool, default: False).
+
+    Return signed data in binary format.
+    """
+
     cmd = "gpg --armor --digest-algo SHA256"
 
     if inline:
@@ -66,14 +78,14 @@ def gpg_sign_string(data, keyname=None, inline=False):
         cmd += " --detach-sign"
 
     if keyname is not None:
-        cmd += " --default-key='%s'" % keyname
+        cmd += " --local-user '%s'" % keyname
 
     proc = subprocess.Popen(cmd,
                             shell=True,
                             stdout=subprocess.PIPE,
                             stdin=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
-    stdout = proc.communicate(input=data)[0]
+    stdout = proc.communicate(input=data.encode('utf-8'))[0]
 
     if proc.returncode != 0:
         raise RuntimeError("Failed to sign file: %s" % stdout)
@@ -82,13 +94,16 @@ def gpg_sign_string(data, keyname=None, inline=False):
 
 
 def sign_metadata(repomdfile):
-    """Requires a proper ~/.rpmmacros file. See <http://fedoranews.org/tchung/gpg/>"""
+    """Requires a proper ~/.rpmmacros file.
+
+    See <http://fedoranews.org/tchung/gpg/>
+    """
     cmd = ["gpg", "--detach-sign", "--armor", "--digest-algo SHA256", repomdfile]
     try:
         subprocess.check_call(cmd)
-        print ("Successfully signed repository metadata file")
-    except subprocess.CalledProcessError as e:
-        print ("Unable to sign repository metadata '%s'" % (repomdfile))
+        print("Successfully signed repository metadata file")
+    except subprocess.CalledProcessError:
+        print("Unable to sign repository metadata '%s'" % repomdfile)
         exit(1)
 
 
@@ -108,10 +123,15 @@ def parse_repomd(data):
     root = ET.fromstring(data)
     namespaces = {'repo': 'http://linux.duke.edu/metadata/repo'}
 
-    filelists = None
-    primary = None
+    filelists = {}
+    primary = {}
+    other = {}
 
-    revision = root.find('repo:revision', namespaces).text
+    # The revision is an optional XML element and may be absent.
+    revision = '0'
+    revision_element = root.find('repo:revision', namespaces)
+    if revision_element:
+        revision = revision_element.text
 
     for child in root:
         if 'type' not in child.attrib:
@@ -128,8 +148,10 @@ def parse_repomd(data):
             filelists = result
         elif child.attrib['type'] == 'primary':
             primary = result
+        elif child.attrib['type'] == 'other':
+            other = result
 
-    return filelists, primary, revision
+    return filelists, primary, other, revision
 
 
 def parse_filelists(data):
@@ -361,12 +383,66 @@ def parse_primary(data):
     return packages
 
 
+def parse_other(data):
+    """Parse other.xml
+
+    Find 'other:changelog' and 'other:version' and
+    fill changelog, version and package.
+
+    Keyword Arguments:
+    data - The unparsed data of other.xml.gz (byte, default: None)
+
+    Return Parsed data as a packages (dict)
+    """
+    root = ET.fromstring(data)
+
+    namespaces = {'other': 'http://linux.duke.edu/metadata/other'}
+
+    packages = {}
+
+    for child in root:
+        if not child.tag.endswith('}package'):
+            continue
+
+        package_id = child.attrib['pkgid']
+        name = child.attrib['name']
+        arch = child.attrib['arch']
+        version = child.find('other:version', namespaces)
+        version = {
+            'ver': version.attrib['ver'],
+            'rel': version.attrib['rel'],
+            'epoch': version.attrib.get('epoch', '0'),
+        }
+
+        changelog_list = child.findall('other:changelog', namespaces)
+
+        changelog = []
+
+        for log in changelog_list:
+            changelog.append({
+                'author': log.attrib['author'],
+                'date': log.attrib['date'],
+                'text': log.text,
+            })
+
+        package = {'pkgid': package_id, 'name': name, 'arch': arch,
+                   'version': version, 'changelog': changelog}
+        nerv = (name, version['epoch'], version['rel'], version['ver'])
+
+        packages[nerv] = package
+
+    return packages
+
+
 def dump_primary(primary):
     res = ""
 
     res += '<?xml version="1.0" encoding="UTF-8"?>\n'
-    res += '<metadata xmlns="http://linux.duke.edu/metadata/common" xmlns:rpm="http://linux.duke.edu/metadata/rpm" packages="%d">\n' % len(
-        primary)
+    res += (
+        '<metadata xmlns="http://linux.duke.edu/metadata/common" '
+        'xmlns:rpm="http://linux.duke.edu/metadata/rpm" '
+        'packages="%d">\n' % len(primary)
+    )
 
     for package in primary.values():
         res += '<package type="rpm">\n'
@@ -379,8 +455,8 @@ def dump_primary(primary):
                                for c in ['epoch', 'ver', 'rel'] if ver[c]])
         res += '%s/>\n' % components
 
-        res += '  <checksum type="sha256" pkgid="YES">%s</checksum>\n' % \
-            package['checksum']
+        res += '  <checksum type="sha256" pkgid="YES">%s</checksum>\n' % (
+            package['checksum'])
 
         res += '  <summary>%s</summary>\n' % escape(package['summary'] or '')
         res += '  <description>%s</description>\n' % escape(
@@ -472,20 +548,135 @@ def parse_ver_str(ver_str):
     epoch = match.group(1)[:-1] if match.group(1) else "0"
     ver = match.group(2)
     rel = match.group(3)[1:] if match.group(3) else None
-    return (epoch, ver, rel)
+    return epoch, ver, rel
 
 
-def get_with_decode(dictionary, key, default='' , encoding='utf-8'):
+def header_to_other(header, sha256):
+    """Method that decodes data for parsing in sha256
+
+    Keyword arguments:
+    header - data that get passed to decode (string, default: None)
+    sha256 - code format that is assigned for generating package id (string)
+
+    Return data from the header of xml tree (name, version, author, etc)
+    """
+    pkgid = sha256
+    name = get_with_decode(header, 'NAME', None)
+    arch = get_arch_from_header(header)
+    epoch = header.get('EPOCH', '0')
+    rel = get_with_decode(header, 'RELEASE', None)
+    ver = get_with_decode(header, 'VERSION', None)
+    version = {'ver': ver, 'rel': rel, 'epoch': epoch}
+
+    package = {
+        'pkgid': pkgid,
+        'name': name,
+        'arch': arch,
+        'version': version,
+        'changelog': [],
+    }
+
+    changelog_name = header.get('CHANGELOGNAME', [])
+    if not isinstance(changelog_name, list):
+        changelog_name = [changelog_name]
+    changelog_text = header.get('CHANGELOGTEXT', [])
+    if not isinstance(changelog_text, list):
+        changelog_text = [changelog_text]
+    changelog_date = header.get('CHANGELOGTIME', [])
+    if not isinstance(changelog_date, list):
+        changelog_date = [changelog_date]
+
+    # get newest CHANGELOG_LIMIT lines in terms of time
+    # (initial commit is oldest), we order these lines as a createrepo does
+    # in reversed order (from old to new)
+    changelog_name = changelog_name[:CHANGELOG_LIMIT][::-1]
+    changelog_text = changelog_text[:CHANGELOG_LIMIT][::-1]
+    changelog_date = changelog_date[:CHANGELOG_LIMIT][::-1]
+
+    for date, author, text in zip(changelog_date, changelog_name, changelog_text):
+        package['changelog'].append({
+            'author': escape(author.decode('utf-8')),
+            'date': date,
+            'text': escape(text.decode('utf-8')),
+        })
+
+    nerv = (name, version['epoch'], version['rel'], version['ver'])
+    return nerv, package
+
+
+def dump_other(other):
+    """Generate other.xml.gz info
+
+    The method generates information for all packages in next structure
+    consequently:
+    <package pkgid="..." name="..." arch="..."
+        <version epoch="..." ver="..." rel="..."/>
+        <changelog author="..." date"...">...</changelog>
+    </package>
+
+    Keyword arguments:
+    other - other data for packages (dict)
+
+    Return full xml tree of an other data (string)
+    """
+    res = ""
+    res += '<?xml version="1.0" encoding="UTF-8"?>\n'
+    res += '<otherdata xmlns="http://linux.duke.edu/metadata/other" packages="%d">\n' % len(other)
+
+    for package in other.values():
+        res += '<package pkgid="%s" name="%s" arch="%s">\n' % (
+            package['pkgid'], package['name'], package['arch'])
+
+        ver = package['version']
+        log = package['changelog']
+
+        res += '  <version '
+        components = ' '.join(
+            ['%s="%s"' % (c, ver[c]) for c in ['epoch', 'ver', 'rel'] if ver[c]])
+        res += '%s/>\n' % components
+
+        for changelog in log:
+            res += '  <changelog author="%s" date="%s">%s</changelog>\n' % (
+                escape(changelog['author']), changelog['date'], escape(changelog['text']))
+
+        res += '</package>\n'
+
+    res += "</otherdata>"
+
+    return res
+
+
+def get_with_decode(dictionary, key, default='', encoding='utf-8'):
     res = dictionary.get(key, default)
     if res:
         res = res.decode(encoding)
     return res
 
 
+def get_arch_from_header(header):
+    """Defines the architecture of the package according to
+    the data from the header.
+
+    Keyword arguments:
+    header - parsed rpm package header (dict)
+
+    Return the architecture the package is for (string)
+    """
+
+    # The architecture definition condition is based on
+    # https://github.com/rpm-software-management/yum/blob/4ed25525ee4781907bd204018c27f44948ed83fe/yum/packages.py#L2222
+    sourcepackage = header.get('SOURCEPACKAGE', None)
+    sourcerpm = get_with_decode(header, 'SOURCERPM', '')
+    if sourcepackage == 1 or not sourcerpm:
+        return 'src'
+    else:
+        return get_with_decode(header, 'ARCH', None)
+
+
 def header_to_filelists(header, sha256):
     pkgid = sha256
     name = get_with_decode(header, 'NAME', None)
-    arch = get_with_decode(header, 'ARCH', None)
+    arch = get_arch_from_header(header)
     epoch = header.get('EPOCH', '0')
     rel = get_with_decode(header, 'RELEASE', None)
     ver = get_with_decode(header, 'VERSION', None)
@@ -541,7 +732,7 @@ def header_to_primary(
         header_end,
         size):
     name = get_with_decode(header, 'NAME', None)
-    arch = get_with_decode(header, 'ARCH')
+    arch = get_arch_from_header(header)
     summary = get_with_decode(header, 'SUMMARY')
     description = get_with_decode(header, 'DESCRIPTION')
     packager = get_with_decode(header, 'PACKAGER', None)
@@ -702,17 +893,24 @@ def header_to_primary(
     return nerv, package
 
 
-def generate_repomd(filelists_str, filelists_gz, primary_str, primary_gz, revision):
+def generate_repomd(filelists_str, filelists_gz,
+                    primary_str, primary_gz,
+                    other_str, other_gz, revision):
     filelists_bytes = filelists_str.encode('utf-8')
     primary_bytes = primary_str.encode('utf-8')
+    other_bytes = other_str.encode('utf-8')
+
     filelists_str_sha256 = bytes_checksum(filelists_bytes, 'sha256')
     primary_str_sha256 = bytes_checksum(primary_bytes, 'sha256')
+    other_str_sha256 = bytes_checksum(other_bytes, 'sha256')
 
     filelists_gz_sha256 = bytes_checksum(filelists_gz, 'sha256')
     primary_gz_sha256 = bytes_checksum(primary_gz, 'sha256')
+    other_gz_sha256 = bytes_checksum(other_gz, 'sha256')
 
     filelists_name = 'repodata/%s-filelists.xml.gz' % filelists_gz_sha256
     primary_name = 'repodata/%s-primary.xml.gz' % primary_gz_sha256
+    other_name = 'repodata/%s-other.xml.gz' % other_gz_sha256
 
     nowdt = datetime.datetime.now()
     nowtuple = nowdt.timetuple()
@@ -721,13 +919,17 @@ def generate_repomd(filelists_str, filelists_gz, primary_str, primary_gz, revisi
     res = ""
 
     res += '<?xml version="1.0" encoding="UTF-8"?>\n'
-    res += '<repomd xmlns="http://linux.duke.edu/metadata/repo" xmlns:rpm="http://linux.duke.edu/metadata/rpm">\n'
+    res += (
+        '<repomd xmlns="http://linux.duke.edu/metadata/repo" '
+        'xmlns:rpm="http://linux.duke.edu/metadata/rpm">\n'
+    )
 
     res += '  <revision>%s</revision>\n' % revision
 
     res += '  <data type="filelists">\n'
     res += '    <checksum type="sha256">%s</checksum>\n' % filelists_gz_sha256
-    res += '    <open-checksum type="sha256">%s</open-checksum>\n' % filelists_str_sha256
+    res += '    <open-checksum type="sha256">%s</open-checksum>\n' % (
+        filelists_str_sha256)
     res += '    <location href="%s"/>\n' % filelists_name
     res += '    <timestamp>%s</timestamp>\n' % int(nowtimestamp)
     res += '    <size>%s</size>\n' % len(filelists_gz)
@@ -736,37 +938,97 @@ def generate_repomd(filelists_str, filelists_gz, primary_str, primary_gz, revisi
 
     res += '  <data type="primary">\n'
     res += '    <checksum type="sha256">%s</checksum>\n' % primary_gz_sha256
-    res += '    <open-checksum type="sha256">%s</open-checksum>\n' % primary_str_sha256
+    res += '    <open-checksum type="sha256">%s</open-checksum>\n' % (
+        primary_str_sha256)
     res += '    <location href="%s"/>\n' % primary_name
     res += '    <timestamp>%s</timestamp>\n' % int(nowtimestamp)
     res += '    <size>%s</size>\n' % len(primary_gz)
     res += '    <open-size>%s</open-size>\n' % len(primary_bytes)
     res += '  </data>\n'
 
+    res += '  <data type="other">\n'
+    res += '    <checksum type="sha256">%s</checksum>\n' % other_gz_sha256
+    res += '    <open-checksum type="sha256">%s</open-checksum>\n' % other_str_sha256
+    res += '    <location href="%s"/>\n' % other_name
+    res += '    <timestamp>%s</timestamp>\n' % int(nowtimestamp)
+    res += '    <size>%s</size>\n' % len(other_gz)
+    res += '    <open-size>%s</open-size>\n' % len(other_bytes)
+    res += '  </data>\n'
+
     res += '</repomd>\n'
 
     return res
 
 
-def update_repo(storage, sign, tempdir):
+def save_malformed_list(storage, malformed_list):
+    """Save the list of malformed packages to the storage.
+
+    Keyword arguments:
+    storage - storage with repositories (Storage object).
+    malformed_list - list of malformed packages (list of strings).
+    """
+    file = 'repodata/malformed_list.txt'
+    if malformed_list:
+        print('Save malformed list...')
+        storage.write_file(file, '\n'.join(malformed_list).encode('utf-8'))
+    elif storage.exists(file):
+        # The list existed before, but is not up-to-date now.
+        print('Delete malformed list...')
+        storage.delete_file(file)
+
+
+def parse_metafiles(storage):
+    """Parse metafiles.
+
+    Keyword arguments:
+    storage - storage with repositories (Storage object).
+    """
     filelists = {}
     primary = {}
+    others = {}
     revision = "0"
     initial_filelists = None
     initial_primary = None
+    initial_others = None
 
     if storage.exists('repodata/repomd.xml'):
         data = storage.read_file('repodata/repomd.xml')
 
-        filelists, primary, revision = parse_repomd(data)
+        filelists, primary, others, revision = parse_repomd(data)
 
-        initial_filelists = filelists['location']
-        data = storage.read_file(initial_filelists)
-        filelists = parse_filelists(gunzip_bytes(data))
+        initial_filelists = filelists.get('location', None)
+        # The file can be specified in repomd.xml but doesn't exist.
+        if initial_filelists and storage.exists(initial_filelists):
+            data = storage.read_file(initial_filelists)
+            filelists = parse_filelists(gunzip_bytes(data))
+        else:
+            initial_filelists = None
+            filelists = {}
 
-        initial_primary = primary['location']
-        data = storage.read_file(initial_primary)
-        primary = parse_primary(gunzip_bytes(data))
+        initial_primary = primary.get('location', None)
+        # The file can be specified in repomd.xml but doesn't exist.
+        if initial_primary and storage.exists(initial_primary):
+            data = storage.read_file(initial_primary)
+            primary = parse_primary(gunzip_bytes(data))
+        else:
+            initial_primary = None
+            primary = {}
+
+        initial_others = others.get('location', None)
+        if initial_others and storage.exists(initial_others):
+            data = storage.read_file(initial_others)
+            others = parse_other(gunzip_bytes(data))
+        else:
+            initial_others = None
+            others = {}
+
+    return (filelists, primary, others, revision,
+            initial_filelists, initial_primary, initial_others)
+
+
+def update_repo(storage, sign, tempdir, force=False):
+    (filelists, primary, others, revision,
+     initial_filelists, initial_primary, initial_other) = parse_metafiles(storage)
 
     recorded_files = set()
     for package in primary.values():
@@ -785,6 +1047,9 @@ def update_repo(storage, sign, tempdir):
         existing_files.add((file_path, mtime))
 
     files_to_add = existing_files - recorded_files
+    # List of packages that can't be added to the index
+    # (some problems encountered during processing).
+    malformed_list = []
 
     for file_to_add in files_to_add:
         file_path = file_to_add[0]
@@ -795,7 +1060,18 @@ def update_repo(storage, sign, tempdir):
         storage.download_file(file_path, os.path.join(tmpdir, 'package.rpm'))
 
         rpminfo = rpmfile.RpmInfo()
-        header = rpminfo.parse_file(os.path.join(tmpdir, 'package.rpm'))
+        header = None
+
+        try:
+            header = rpminfo.parse_file(os.path.join(tmpdir, 'package.rpm'))
+        except Exception as err:
+            print("Can't parse '%s':\n%s" % (file_path, str(err)))
+            if force:
+                malformed_list.append(file_path)
+                continue
+            else:
+                raise err
+
         sha256 = file_checksum(os.path.join(tmpdir, 'package.rpm'), "sha256")
 
         statinfo = os.stat(os.path.join(tmpdir, 'package.rpm'))
@@ -807,38 +1083,61 @@ def update_repo(storage, sign, tempdir):
                                        rpminfo.header_start, rpminfo.header_end,
                                        size)
         _, flist = header_to_filelists(header, sha256)
+        _, other = header_to_other(header, sha256)
 
         primary[nerv] = prim
         filelists[nerv] = flist
+        others[nerv] = other
+
+    save_malformed_list(storage, malformed_list)
 
     revision = str(int(revision) + 1)
 
     filelists_str = dump_filelists(filelists)
     primary_str = dump_primary(primary)
+    other_str = dump_other(others)
+
     filelists_gz = gzip_bytes(filelists_str.encode('utf-8'))
     primary_gz = gzip_bytes(primary_str.encode('utf-8'))
+    other_gz = gzip_bytes(other_str.encode('utf-8'))
 
     repomd_str = generate_repomd(filelists_str, filelists_gz,
-                                 primary_str, primary_gz, revision)
+                                 primary_str, primary_gz,
+                                 other_str, other_gz, revision)
 
     filelists_gz_sha256 = bytes_checksum(filelists_gz, 'sha256')
     primary_gz_sha256 = bytes_checksum(primary_gz, 'sha256')
+    other_gz_sha256 = bytes_checksum(other_gz, 'sha256')
+
     filelists_name = 'repodata/%s-filelists.xml.gz' % filelists_gz_sha256
     primary_name = 'repodata/%s-primary.xml.gz' % primary_gz_sha256
+    other_name = 'repodata/%s-other.xml.gz' % other_gz_sha256
 
     storage.write_file(filelists_name, filelists_gz)
     storage.write_file(primary_name, primary_gz)
+    storage.write_file(other_name, other_gz)
     storage.write_file('repodata/repomd.xml', repomd_str.encode('utf-8'))
 
-    if initial_filelists:
+    # Here we are deleting few old metafiles.
+    # The difference in names between the old and new files (in case the new
+    # packages were not added to the repository) exists because part of the
+    # name is the sha256 hash from  the".gz" files, wich includes information
+    # about time when it was created (in seconds (float)). In my opinion, this
+    # difference is not obvious on the one hand and may cease to exist on the
+    # other hand (if we start to set timestamp according to a different logic
+    # or if the hashsum is removed from the filename).
+    # Let's check the names for equivalence.
+    if initial_filelists and initial_filelists != filelists_name:
         storage.delete_file(initial_filelists)
-    if initial_primary:
+    if initial_primary and initial_primary != primary_name:
         storage.delete_file(initial_primary)
+    if initial_other and initial_other != other_name:
+        storage.delete_file(initial_other)
 
     if sign:
-        repomd_str_signed = gpg_sign_string(repomd_str)
-        storage.write_file('repodata/repomd.xml.asc',
-            repomd_str_signed.encode('utf-8'))
+        keyname = os.getenv('GPG_SIGN_KEY')
+        repomd_signed = gpg_sign_string(repomd_str, keyname)
+        storage.write_file('repodata/repomd.xml.asc', repomd_signed)
 
 
 def main():
